/*
  Etripator: PCE disassembler
  Author   : Vincent Cruz
  Date     : 2004-End of the world
  [include licence here]
  i'm not responsible of any damage this program wil cause.
  use it on your own risk. blablabla...
*/
#include <stdio.h>
#include <stdlib.h>
#include <csv.h>
#include <section.h>
#include <opcodes.h>
#include <getopt.h>

char* IRQNames[5] = {
	"irq_2",
	"irq_1",
	"irq_timer",
	"irq_nmi",
	"irq_reset"
};

/* General options */
int g_mkit_output = 0;

/*
  Process opcode
*/
int process_opcode(unsigned char* i_file, unsigned int* i_offset, char* i_eor, FILE* i_out) {
  int i;
  unsigned char inst;
  unsigned char data[6];

  /* Opcode */
  inst = *(i_file++);

  /* End Of Routine (eor) is set to 1 if the instruction is RTI or RTS */
  if((inst == 64) || (inst == 96)) *i_eor = 1;

  if(!g_mkit_output)
  {
    /* Print offset and opcode value */
    fprintf(i_out, "%04x - %02x ",*i_offset, inst);
  }

  /* Data */
  if(pce_opcode[inst].size > 1) {
    for(i=1; i<pce_opcode[inst].size; ++i) {
      data[i-1] = *(i_file++);
    }
    if(!g_mkit_output)
    {
      for(i=1; i<pce_opcode[inst].size; ++i) {
        fprintf(i_out, "%02x ",data[i-1]);
      }
    }
  
    /* Swap LSB and MSB (woo!) */
    if(pce_opcode[inst].size > 2)
    {
  	  unsigned char swap;
	  for(i=0; i<(pce_opcode[inst].size-1); i+=2)
	  {
	    swap      = data[i];
	    data[i]   = data[i+1];
	    data[i+1] = swap;
	  }
    }   
  }
  
  if(!g_mkit_output)
  {
    for(i=pce_opcode[inst].size; i<7; ++i) {
      fprintf(i_out, "   ");
    }
  }
  /*
    Warning:
    BBRi, BBSi, BCC, BCS, BEQ, BMI, BNE, BPL, BRA, BSR, BVC, BVS:
      OPC hhll     | 
      OPC ZZ, hhll | hhll = data + current program offset
  */  
  if((pce_opcode[inst].type == 13) || ((pce_opcode[inst].size == 2) &&
	 (pce_opcode[inst].type == 14))) {
    unsigned short int tmp;
    signed char delta;

    i = (pce_opcode[inst].type == 13)?1:0;
   
    /* Detect negative number */
    if(data[i] & 128)
    	delta = - ((data[i] - 1) ^ 0xff);
    else
    	delta = data[i];
    
    tmp = *i_offset + delta + pce_opcode[inst].size;
    data[i  ] = (tmp >> 8) & 0xff;
    data[i+1] = tmp & 0xff;
  }

  *i_offset += pce_opcode[inst].size;

  /* Print opcode sting */
  fprintf(i_out,"%c %s ", g_mkit_output ? ' ' : '-', pce_opcode[inst].name);
  
  /* Accumulator? */
  if(pce_opcode[inst].type == 1) {
    fprintf(i_out,"A");
  }
  else {
    /* Print data */
    if(pce_opcode[inst].type) {
      for(i=0; pce_opstring[pce_opcode[inst].type][i] != NULL; ++i) {
			fprintf(i_out, pce_opstring[pce_opcode[inst].type][i], data[i]);
      }
    }
  }

  fprintf(i_out,"\n");

  return pce_opcode[inst].size;
}

void usage()
{
	fprintf(stderr, "Usage : etripator [options] <csv> <in>\n"
					"-m , --mkit : output using mkit syntax\n");
}

/*
  TODO :
      Data detection
*/
int main(int argc, char** argv) {
  FILE* out;
  FILE* in;

  char eor;

  unsigned char* rom;
  unsigned char* ptr;

  unsigned int size;
  
  unsigned int offset;

  unsigned int i, j;

  int idx, opt;

  Section* section;
  size_t sectionCount;

  char cmd_line_options[] = "m";
  struct option cmd_line_long_options[] = {
	{"mkit",    0, &g_mkit_output,  1 },
	{     0,	0,              0,  0 }
  };

  while ((opt = getopt_long (argc, argv, cmd_line_options, cmd_line_long_options, &idx)) > 0)
  {
	switch(opt)
	{
		case 'm':
			g_mkit_output = 1;
			break;

		case 'h':
			usage();
			return 0;
			
		default:
			usage();
			return 1;
	}		
  }

  if((argc - optind) != 2)
  {
	usage();
	exit(1);
  }

  /* Read csv file */
  sectionCount = 5;
  section      = NULL;
  if(!readSectionsFromCSV(argv[optind],';',&section, &sectionCount)) {
  	fprintf(stderr,"Unable to read %s. Only irq will be desassembled.\n", argv[optind]);
  	if(section) free(section);
  	sectionCount = 5;
  	section = (Section*)malloc(sectionCount * sizeof(Section));
  }

  in = fopen(argv[optind+1], "rb");
  if(!in) {
	fprintf(stderr,"Can't open %s\n", argv[optind+1]);
	exit(1);
  }  

  fseek(in, 0, SEEK_END);

  /* Get file size */
  size  = ftell(in);
  fseek(in, 0, SEEK_SET);
  size -= ftell(in);

  /* Adjust if header present */

  fseek(in, (size & 0x1fff), SEEK_SET);
  size &= ~0x1fff;

  rom = (unsigned char*)malloc(size * sizeof(unsigned char));
  
  fread(rom, 1, size, in);

  fclose(in);

  /* Output irq offsets */
  ptr = rom + 0x1ff6;
  for(i=0; i<5; ++i) {
  	offset = ((*(ptr+1) << 8) | *ptr) - 0xe000;
  	printf("irq #%d : %04x\n", i, offset);

	section[i].name  = IRQNames[i];
  	section[i].type  = CODE;
  	section[i].start = offset;
  	section[i].end   = -1;
  	
  	ptr += 2;
  }

  /* Disassemble and output */
  for(i=0; i<sectionCount; ++i) {
	out = fopen(section[i].name, "wb");
	if(!out) {
		fprintf(stderr,"Can't open %s\n", section[i].name);
		exit(1);
	}

	ptr    = rom + section[i].start;
	offset = section[i].start;
	eor = 0;
			
	if(section[i].type == CODE) {
		if(g_mkit_output)
		{
			fprintf(out, "\t.code\n\t.bank 0         ;FIXME\n\t.org $%04lx      ;FIXME\n",
						section[i].start & 8192);
		}
		
		while(!eor) {
			ptr += process_opcode(ptr, &offset, &eor, out);
			
			if((section[i].end > 0) && (offset >= section[i].end)) {
				eor = 1;
			}
		}
	}
	else {
		if(section[i].end <= 0) section[i].end = size;
		
		if(g_mkit_output)
		{
			fprintf(out, "\t.data\n\t.bank 0         ;FIXME\n\t.org $%04lx      ;FIXME\n",
						section[i].start & 8192);
						
			while(!eor) {
				fprintf(out, "  .db ");
				
				
				j=0;
				while(1)
				{
					fprintf(out, "$%02x", rom[offset]);

					++j;
					++offset;
										
					if((j<8) && (offset < section[i].end))
					{
						fprintf(out, ", ");
					}
					else
					{
						fprintf(out, "\n");
						break;
					}
				}

				if((section[i].end > 0) && (offset >= section[i].end)) {
					eor  = 1;
				}
			}
		}
		else
		{
			fwrite(rom + offset, section[i].end - section[i].start, 1, out);
		}
	}
	
	fclose(out);
  }
  free(rom);

  for(i=5; i<sectionCount; ++i) {
  	free(section[i].name);
  	section[i].name = NULL;
  }
  free(section);

  return 0;
}
